// Jakub Sitnicki, grupa F1IBERE

/*
 * Demonstracja biblioteki boost::iostreams - acuchy filtrw
 * (komentarze w stylu C zostay uyte celowo)
 *
 * Kompilacja:
 * gcc -Wall -pedantic -lboost_iostreams boost_iostreams_filter_chains.cpp
 */

#include <cstdio>		/* EOF */
#include <iostream>		/* std::cout */
#include <istream>		/* std::istream */

#include <boost/iostreams/device/file_descriptor.hpp>
#include <boost/iostreams/device/file.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filter/line.hpp>
#include <boost/iostreams/filter/stdio.hpp>

/*
 * Aby zademonstrowa uycie acuchw filtrw, musimy najpierw
 * utworzy kilka prostych filtrw, ktre posu nam do tego
 * celu. Tworzenie filtrw nie ley w zakresie tej demonstracji, wic
 * zostan one jedynie krtko przedstawione
 */

/*
 * Filtr zastpujcy znaki wystpujce w pierwszym zbiorze (set1),
 * odpowiadajcymi im znakami z drugiego zbioru (set2). Jeeli znak
 * wystpujcy w pierwszym zbiorze nie ma odpowiednika w drugim
 * zbiorze to zostanie on usunity ze strumienia. Filtr w swoim
 * dziaaniu przypomina narzdzie powoki tr(1).
 */
class translate_chars_filter : public boost::iostreams::stdio_filter {
public:
	explicit translate_chars_filter(const std::string& set1, const std::string& set2 = std::string())
		: set1_(set1), set2_(set2)
		{ }
private:
	std::string set1_;
	std::string set2_;

	void do_filter() {
		int c;
		while ((c = std::cin.get()) != EOF) {
			std::string::size_type pos = set1_.find(c);
			if (pos != std::string::npos) {
				try {
					/* zamie znak */
					c = set2_.at(pos);
				} catch (std::out_of_range e) {
					/* usu znak */
					continue;
				}
			}

			std::cout.put(c);
		}
	}
};
BOOST_IOSTREAMS_PIPABLE(translate_chars_filter, 0)

/*
 * Filtr zastpujcy wystpienie jednego acucha znakw w strumieniu
 * innym acuchem znakw.
 */
class replace_string_filter : public boost::iostreams::stdio_filter {
public:
	explicit replace_string_filter(const std::string& search,
				       const std::string& replace)
		: search_(search), replace_(replace)
		{ }
private:
	std::string search_;
	std::string replace_;

	void do_filter() {
		int c;
		std::string buf;

		while ((c = std::cin.get()) != EOF) {
			if (buf.size() == search_.size()) {
				std::cout << *buf.begin();
				buf.erase(buf.begin());
			}

			buf.push_back(c);

			if (buf == search_) {
				std::cout << replace_;
				buf.clear();
			}
		}

		std::cout << buf;
	}
};
BOOST_IOSTREAMS_PIPABLE(replace_string_filter, 0)

/*
 * Filtr usuwajcy parzyste linie ze strumienia.
 */
class skip_even_lines_filter : public boost::iostreams::line_filter {
public:

	explicit skip_even_lines_filter() : odd_(true) { }
private:
	bool odd_;

	std::string do_filter(const std::string& line) {
		if (odd_) {
			odd_ = false;
			return line;
		} else {
			odd_ = true;
			return std::string();
		}
	}
};
BOOST_IOSTREAMS_PIPABLE(skip_even_lines_filter, 0)

/*
 * W celu zademonstrowania uycia acuchw filtrw w wymylny sposb
 * usuniemy wszystkie komentarze w stylu C z pliku
 * wejciowego. Uyjemy nastpujcego algorytmu:
 *
 * 1. zastp znaczniki pocztku i koca komentarza znakiem
 *    '\377' (0xFF),
 * 2. zamie znaki nowej linii ('\n') na znaki '\377', natomiast znaki
 *    '\377' na znaki nowej linii ('\n'),
 * 3. usu co drug lini (czyli wszystkie parzyste),
 * 4. usu znaki nowej linii,
 * 5. zastp znaki '\377' znakami nowej linii.
 *
 * Plikiem wejciowym bdzie ten oto plik rdowy.
 */
int main()
{
	using namespace std;
	using namespace boost;

	/*
	 * O acuchach filtrw moemy myle jako o stosie
	 * filtrw. acuch filtrw jest reprezentowany przez klas
	 * iostreams::chain, ktrej nie uywamy bezporednio.
	 *
	 * acuchy filtrw stosujemy gwnie w poczeniu z
	 * filtrowanymi strumieniami - iostream::filtering_stream oraz
	 * iostream::filtering_streambuf.
	 */

	/*
	 * W przypadku filtrowanego strumienia _wejciowego_, kolejne
	 * filtry dodajemy do acucha w kolejnoci odwrotnej ni
	 * bdzie przez nie przepywa strumie danych. (Tzn. ostatni
	 * filtr dodajemy jako pierwszy.) acuch wejciowy musi
	 * koczy si rdem (Source), ktre moe by
	 * np. strumieniem wejciowym std::istream.
	 */
	iostreams::filtering_istream in1;

	/* dodajemy filtry zgodnie z omwionym algorytmem w odwrotnej
	 * kolejnoci */
	in1.push(translate_chars_filter("\377", "\n")); /* krok 5 */
	in1.push(translate_chars_filter("\n")); /* krok 4 */
	in1.push(skip_even_lines_filter()); /* krok 3 */
	in1.push(translate_chars_filter("\n\377", "\377\n")); /* krok 2 */
	in1.push(replace_string_filter("*/", "\377")); /* krok 1 cd. */
	in1.push(replace_string_filter("/*", "\377")); /* krok 1 */
	in1.push(iostreams::file_descriptor_source(__FILE__)); /* rdo */

	/* aby uy filtru przepisujemy filtrowany strumie wejciowy
	 * na wyjcie; moemy z niego czyta korzystajc z interfejsu
	 * std::istream */
 	std::cout << in1.rdbuf();


	/*
	 * Jeeli filtru zostay zdefiniowane przy uyciu makra
	 * BOOST_IOSTREAMS_PIPABLE, to acuch moemy rwnie
	 * zdefiniowa uywajc przecionych operatorw | (skadnia
	 * podobna do potokw w powoce):
	 * filtr1 | filtr2 | ... | rdo
	 */
	iostreams::filtering_istream in2(
		/* krok 5 i 4 */
		translate_chars_filter("\377", "\n") | translate_chars_filter("\n") |
		/* krok 3 i 2 */
		skip_even_lines_filter() | translate_chars_filter("\n\377", "\377\n") |
		/* krok 1 */
		replace_string_filter("*/", "\377") | replace_string_filter("/*", "\377") |
		/* rdo */
		iostreams::file_descriptor_source(__FILE__)
		);

	/* (usu komentarz aby obejrze rezultat powyszej filtracji) */
	//std::cout << in2.rdbuf();

	/*
	 * Definiujc acuch filtrw dla strumienia _wyjciowego_,
	 * filtry podajemy w takiej kolejnoci w jakiej ma przez nie
	 * przepywa strumie. acuch musi by zakoczony
	 * zbiornikiem (Sink), np. strumieniem wyjciowym
	 * std::ostream.
	 */
	iostreams::filtering_ostream out;

	out.push(replace_string_filter("/*", "\377")); /* krok 1 */
	out.push(replace_string_filter("*/", "\377")); /* krok 1 cd. */
	out.push(translate_chars_filter("\n\377", "\377\n")); /* krok 2 */
	out.push(skip_even_lines_filter()); /* krok 3 */
	out.push(translate_chars_filter("\n")); /* krok 4 */
	out.push(translate_chars_filter("\377", "\n")); /* krok 5 */
	out.push(std::cout); /* zbiornik */

	/* rdo, ktrego zawarto bdziemy przetwarza */
	iostreams::stream_buffer<iostreams::file_source> buf(__FILE__);
	std::istream fin(&buf);

	/* (usu komentarz aby obejrze rezultat powyszej filtracji) */
	//out << fin.rdbuf();

	return 0;
}

/*
 * Demonstracja zostaa zainspirowana skryptem powoki pochodzcym z
 * przewodnika "Advanced Bash-Scripting Guide":
 * http://tldp.org/LDP/abs/html/filearchiv.html#STRIPC
 */
